# coding: utf-8

"""

    Open-i Text Feature Extraction
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
import re
import string
from bs4 import BeautifulSoup
from collections import defaultdict

# Image Support Tools
from biovida.images._interface_support.openi.openi_support_tools import (filter_unnest,
                                                                         num_word_to_int,
                                                                         multiple_decimal_remove)

# General Support Tools
from biovida.support_tools.support_tools import (cln,
                                                 unescape,
                                                 order_set,
                                                 multi_replace,
                                                 remove_html_bullet_points)

# Data
from biovida.images._interface_support.openi._openi_imaging_modality_information import (terms_dict,
                                                                                         modality_subtypes,
                                                                                         contradictions,
                                                                                         modality_specific_subtypes)


CLINICAL_ARTICLE_TYPES = ('encounter', 'case_report', 'radiology_report')


# ----------------------------------------------------------------------------------------------------------
# Abstract Processing
# ----------------------------------------------------------------------------------------------------------


def _key_clean(key):
    """

    Cleans the key for the ``parsed_abstract`` dictionary inside ``_abstract_parser()``.

    :param key: ``key`` as evolved inside ``_abstract_parser()``.
    :type key: ``str``
    :return: see description.
    :rtype: ``str``
    """
    key_lowered = key.lower()
    if 'history' in key_lowered or 'background' in key_lowered:
        return 'background'
    if 'material' in key_lowered or 'method' in key_lowered:
        return "methods"
    if 'finding' in key_lowered or 'result' in key_lowered:
        return 'results'
    else:
        return cln(cln(key).replace(":", "")).replace("/", "_").replace(" ", "_").lower()


def _value_clean(value):
    """

    Cleans the value for the ``parsed_abstract`` dictionary inside ``_abstract_parser()``.

    :param value: ``value`` as evolved inside ``_abstract_parser()``.
    :type value: ``str``
    :return: see description.
    :rtype: ``str``
    """
    nans = ('na', 'n/a', 'nan', 'none', '')
    value = cln(unescape(value).strip(";"))
    return value if value.lower() not in nans else None


def _abstract_parser(abstract):
    """

    Take a raw HTML abstract and generate dictionary
    using the items in bold (e.g., '<b>History: </b>') as the keys
    and the result of the information inside the <p>...</p> tags as the value.

    :param abstract: a text abstract.
    :type abstract: ``str``
    :return: a dictionary of the form described above.
    :rtype: ``None`` or ``dict``

    :Example:

    >>> _abstract_parser(abstract='<p><b>History: </b>The patient presented with xyz.</p>')
    ...
    {'history': 'The patient presented with xyz.'}

    """
    if not isinstance(abstract, str):
        return None

    soup = BeautifulSoup(remove_html_bullet_points(abstract), 'lxml')

    parsed_abstract = dict()
    for p in soup.find_all('p'):
        # Look for Key
        key_candidate = p.find_all('b')
        if len(key_candidate) == 1:
            key = _key_clean(key=key_candidate[0].text)
            if len(key):
                # Look for Value
                contents = p.contents
                if len(contents) == 2:
                    parsed_abstract[key] = _value_clean(value=contents[-1])

    return parsed_abstract if parsed_abstract else None


# ----------------------------------------------------------------------------------------------------------
# Patient's Sex
# ----------------------------------------------------------------------------------------------------------


def _patient_sex_extract(image_summary_info):
    """

    Tool to Extract the age of a patient.

    :param image_summary_info: some summary text of the image, e.g., 'history', 'abstract', 'image_caption'
                               or 'image_mention'. (Expected to be lower case)
    :type image_summary_info: ``str``
    :return: the sex of the patient.
    :rtype: ``str`` or ``None``
    """
    counts_dict_f = {t: image_summary_info.count(t) for t in ['female', 'woman', 'girl', ' she ', ' f ']}
    counts_dict_m = {t: image_summary_info.count(t) for t in ['male', 'man', 'boy', ' he ', ' m ']}

    # Block Conflicting Information
    if sum(counts_dict_f.values()) > 0 and sum([v for k, v in counts_dict_m.items() if k not in ['male', 'man']]) > 0:
        return None

    # Check for sex information
    if any(x > 0 for x in counts_dict_f.values()):
        return 'female'
    elif any(x > 0 for x in counts_dict_m.values()):
        return 'male'
    else:
        return None


def _patient_sex_guess(background, abstract, image_caption, image_mention):
    """

    Tool to extract the sex of the patient (female or male).

    :param background: the background of the patient/study.
    :type background: ``str``
    :param abstract: a text abstract.
    :type abstract: ``str``
    :param image_caption: an element from the 'image_caption' column.
    :type image_caption: ``str``
    :param image_mention: an element from the 'image_mention' column.
    :type image_mention: ``str``
    :return: the sex of the patent.
    :rtype: ``str`` or ``None``
    """
    for source in (background, abstract, image_caption, image_mention):
        if isinstance(source, str):
            extract = _patient_sex_extract(source.lower())
            if isinstance(extract, str):
                return extract
    else:
        return None


# ----------------------------------------------------------------------------------------------------------
# Patient's Age
# ----------------------------------------------------------------------------------------------------------


def _age_refine(age_list, upper_age_bound=130):
    """

    Converts ages into floats and ages given in months to years.

    :param age_list: the list of ages evolved inside ``_patient_age_guess()``.
    :type age_list: ``list``
    :param upper_age_bound: max. valid age. Defaults to `130`.
    :type upper_age_bound: ``float`` or ``int``
    :return: a numeric age.
    :rtype: ``float`` or ``None``
    """
    to_return = list()
    for a in age_list:
        age = multiple_decimal_remove(a)
        if age is None:
            return None
        try:
            if 'month' in a:
                to_return.append(round(float(age) / 12, 2))
            else:
                to_return.append(float(age))
        except:
            return None

    # Remove Invalid Ages
    valid_ages = [i for i in to_return if i <= upper_age_bound]

    # Heuristic: typically the largest value will be the age
    if len(valid_ages):
        return max(to_return)
    else:
        return None


def _patient_age_guess_abstract_clean(abstract):
    """

    Cleans the abstract for ``_patient_age_guess()``.
    This includes:

    - removing confusing references to the duration of the illness

    - numeric values from 'one' to 'one hundred and thirty' in natural language.

    :param abstract: a text abstract.
    :type abstract: ``str``
    :return: a cleaned ``abstract``
    :rtype: ``str``
    """
    # E.g., '...eight year history of...' --> '...8 year history of...'
    abstract_int = num_word_to_int(abstract)

    # Block: 'x year history'
    history_block = ("year history", " year history", "-year history")

    # Flatten list comp.
    hist_matches = filter_unnest([re.findall(r'\d*\.?\d+' + drop, abstract_int) for drop in history_block])

    if len(hist_matches):
        return cln(" ".join([abstract_int.replace(r, "") for r in hist_matches]))
    else:
        return cln(abstract_int)


def _age_marker_match(image_summary_info):
    """

    Extract age from a string based on it being followed by one of
    those in ``age_markers``.

    Note: will be confused by strings like: 'The first six years of substantial'...

    :param image_summary_info: some summary text of the image, e.g., 'history', 'abstract', 'image_caption'
                               or 'image_mention'.
    :type image_summary_info: ``str``
    :return: patient age.
    :rtype: ``str`` or ``None``
    """
    # Note: given " y", it is not clear how much help something like " yo" is, given the former
    # will already catch it.
    age_markers = (" y", "yo ", " yo ", "y.o.", "y/o", "year", "-year", " -year", " ans ",
                   "month old", " month old", "-month old", "months old", " months old", "-months old")

    # Clean the input text
    cleaned_input = _patient_age_guess_abstract_clean(image_summary_info).replace(" - ", "-").lower()

    # Check abstract
    if len(cleaned_input):
        rslts = filter_unnest([re.findall(r'\d*\.\d+' + b + '|\d+' + b, cleaned_input) for b in age_markers])
        return rslts if len(rslts) else None
    else:
        return None


def _patient_age_guess(background, abstract, image_caption, image_mention):
    """

    Guess the age of the patient.

    :param background: the background of the patient/study.
    :type background: ``str``
    :param abstract: a text abstract.
    :type abstract: ``str``
    :param image_caption: an element from the 'image_caption' column.
    :type image_caption: ``str``
    :param image_mention: an element from the 'image_mention' column.
    :type image_mention: ``str``
    :return: patient age.
    :rtype: ``float`` or ``None``
    """
    # Loop through the inputs, search all of them for age matches
    for source in (background, abstract, image_caption, image_mention):
        if isinstance(source, str):
            matches = _age_marker_match(source)
            if isinstance(matches, list) and len(matches):
                return _age_refine(matches)
    else:
        return None


# ----------------------------------------------------------------------------------------------------------
# Patient's Ethnicity
# ----------------------------------------------------------------------------------------------------------


def _ethnicity_guess_engine(image_summary_info):
    """

    Engine to power ``_ethnicity_guess()``.

    :param image_summary_info: some summary text of the image, e.g., 'history', 'abstract', 'image_caption' or 'image_mention'.
    :type image_summary_info: ``str``
    :return: tuple of the form ('ethnicity', 'sex'), ('ethnicity', None) or (None, None).
    :type: ``tuple``
    """
    e_matches, s_matches = set(), set()

    # Clean the input
    image_summary_info_cln = cln(image_summary_info)

    def eth(ethnicity):
        return ["{0} {1}".format(ethnicity, s) for s in ['adult', 'male', 'female']]

    # Define long form of references to ethnicity
    long_form_ethnicities = {
        'caucasian': ['caucasian'] + eth('white'),
        'black': ['african american'] + eth('black'),
        'latino': ['latino'],
        'asian': ['asian'],
        'native american': ['native american'],
        'pacific islander': ['pacific islander'],
        'first nations': ['first nations'],
        'aboriginal': ['aboriginal']
    }

    for k, v in long_form_ethnicities.items():
        for j in v:
            if j in image_summary_info_cln.lower():
                if not ('caucasian' in e_matches and j == 'asian'):  # 'asian' is a substring in 'caucasian'.
                    e_matches.add(k)

    # Define short form of references to ethnicity. ToDo: use regex to anchor text.
    short_form_ethnicities = [(' BM ', 'black', 'male'), (' BF ', 'black', 'female'),
                              (' WM ', 'caucasian', 'male'), (' WF ', 'caucasian', 'female')]

    for (abbrev, eth, sex) in short_form_ethnicities:
        if abbrev in image_summary_info_cln:
            e_matches.add(eth)
            s_matches.add(sex)

    patient_ethnicity = list(e_matches)[0] if len(e_matches) == 1 else None
    patient_sex = list(s_matches)[0] if len(s_matches) == 1 else None

    return patient_ethnicity, patient_sex


def _ethnicity_guess(background, abstract, image_caption, image_mention):
    """

    Guess the ethnicity of the patient.

    :param background: the background of the patient/study.
    :type background: ``str``
    :param abstract: a text abstract.
    :type abstract: ``str``
    :param image_caption: an element from the 'image_caption' column.
    :type image_caption: ``str``
    :param image_mention: an element from the 'image_mention' column.
    :type image_mention: ``str``
    :return: tuple of the form ('ethnicity', 'sex'), ('ethnicity', None) or (None, None).
    :rtype: ``tuple``
    """
    matches = list()
    for source in (background, abstract, image_caption, image_mention):
        if isinstance(source, str):
            matches.append(_ethnicity_guess_engine(source))

    # 1. Prefer both pieces of information
    both_matches = [i for i in matches if not all(j is None for j in i)]
    if len(both_matches):
        return both_matches[0]  # simply use the first one

    # 2. Search for single matches
    single_matches = [i for i in matches if any(j is not None for j in i)]
    if len(single_matches):
        return single_matches[0]  # again, simply use the first one
    else:
        return None, None


# ----------------------------------------------------------------------------------------------------------
# Patient's Disease
# ----------------------------------------------------------------------------------------------------------


def _remove_substrings(list_of_terms):
    """

    Remove list elements that are substrings of other list elements.

    :param list_of_terms: any list of strings and their starting position in the source string.
    :type list_of_terms: ``iterable``
    :return: see description.
    :type: ``list_of_terms`` with any substrings of other terms removed.

    :Example:

    # Source string: "The patient presented with congenital heart disease...".
    >>> remove_substrings(list_of_terms=[[38, 'heart disease'], [27, 'congenital heart disease'], [44, 'disease']])
    ...
    [27, 'congenital heart disease']

    """
    if len(list_of_terms) <= 1:
        return list_of_terms
    else:
        rslt = list()
        for i in list_of_terms:
            if not any(i[-1] in j[-1] for j in list_of_terms if j[-1] != i[-1]):
                rslt.append(i)
        return rslt


def _disease_guess(problems, title, background, abstract, image_caption, image_mention, list_of_diseases):
    """

    Search `problems`, `title`, `abstract`, `image_caption` and `image_mention` for diseases in `list_of_diseases`

    Near future: count the number of times the disease is mentioned in all sources and use that to guide decision...

    Future Directions:

        Shift to more advanced approaches and consider using one or all of:

        - the NegEx Algorithm (https://healthinformatics.wikispaces.com/NegEx+Algorithm).
        - pyConTextNLP library (https://github.com/chapmanbe/pyConTextNLP) -- more recent tool than NegEx.
        - consider using the 'image_caption_concepts' and mesh column.

    Also see:
        - Interleaved Text/Image Deep Mining on a Large-Scale Radiology Database for Automated Image Interpretation
          (https://arxiv.org/abs/1505.00670).
        - https://irp.nih.gov/blog/post/2016/11/challenges-to-training-artificial-intelligence-with-medical-imaging-data

    :param problems:
    :param title:
    :param background:
    :param abstract:
    :param image_caption:
    :param image_mention:
    :param list_of_diseases: see ``feature_extract``.
    :type list_of_diseases: ``list``
    :return:
    :rtype: ``str`` or ``None``

    :Example:

    >>> from biovida.diagnostics.disease_ont_interface import DiseaseOntInterface
    >>> list_of_diseases = DiseaseOntInterface().pull()['name'].tolist()

    >>> abstract = '...a patient with pancreatitis, not Macrolipasemia' 
    >>> _disease_guess(problems=None, title=None, background=None,
    ...                abstract=abstract, image_caption=None, image_mention=None,
    ...                list_of_diseases=list_of_diseases)
    ...
    'pancreatitis'

    """
    # ToDO: this function is very computationally expensive and rather ineffective. Replace.
    generic = ('syndrome', 'disease')

    if isinstance(problems, str) and cln(problems).lower() in ('normal', 'none'):
        return 'normal'

    possible_diseases = list()
    for source in (image_caption, image_mention, problems, title, background, abstract):
        if isinstance(source, str) and len(source):
            source_clean = cln(BeautifulSoup(source, 'lxml').text.lower())
            for d in list_of_diseases:
                if d in source_clean:
                    possible_diseases.append([source_clean.find(d), d])

    no_substrings = _remove_substrings(possible_diseases)
    to_return = list(set([i[1] for i in no_substrings if i[1] not in generic]))

    if len(to_return):
        return "; ".join(sorted(to_return))
    # Try to fall back to `problems`.
    elif not len(to_return) and isinstance(problems, str):
        filtered_problems = [p for p in problems.split(";") if p in list_of_diseases]
        if len(filtered_problems):
            return "; ".join(sorted(filtered_problems))
        else:
            return None
    else:
        return None


# ----------------------------------------------------------------------------------------------------------
# Illness Duration
# ----------------------------------------------------------------------------------------------------------


def _illness_duration_guess_engine(image_summary_info):
    """

    Engine to search through the possible ways illness duration information could be represented.

    :param image_summary_info: some summary text of the image, e.g., 'history', 'abstract', 'image_caption'
                                or 'image_mention'.
    :type image_summary_info: ``str``
    :return: the best guess for the length of the illness.
    :rtype: ``float`` or ``None``.
    """
    cleaned_source = cln(image_summary_info.replace("-", " "), extent=1)

    match_terms = [('month history of', 'm'), ('year history of', 'y')]
    hist_matches = [(re.findall(r"\d*\.?\d+ (?=" + t + ")", cleaned_source), u) for (t, u) in match_terms]

    def time_adj(mt):
        if len(mt[0]) == 1:
            cleaned_date = re.sub("[^0-9.]", "", cln(mt[0][0], extent=2)).strip()
            if cleaned_date.count(".") > 1:
                return None
            if mt[1] == 'y':
                return float(cleaned_date)
            elif mt[1] == 'm':
                return float(cleaned_date) / 12.0
        else:
            return None

    # Filter invalid extracts
    durations = list(filter(None, map(time_adj, hist_matches)))

    return durations[0] if len(durations) == 1 else None


def _illness_duration_guess(background, abstract, image_caption, image_mention):
    """

    Guess the duration of an illness. Unit: years.

    :param background: the background of the patient/study.
    :type background: ``str``
    :param abstract: a text abstract.
    :type abstract: ``str``
    :param image_caption: an element from the 'image_caption' column.
    :type image_caption: ``str``
    :param image_mention: an element from the 'image_mention' column.
    :type image_mention: ``str``
    :return:
    :rtype: ``float`` or ``None``
    """
    for source in (background, abstract, image_caption, image_mention):
        if isinstance(source, str):
            duration_guess = _illness_duration_guess_engine(source.lower())
            if isinstance(duration_guess, float):
                return duration_guess
    else:
        return None


# ----------------------------------------------------------------------------------------------------------
# Imaging Technology (Modality)
# ----------------------------------------------------------------------------------------------------------


def _im_scan(source):
    """

    Scan a ``source`` for modality information.

    :param source:
    :return:
    """
    # 1. Look for modality matches
    matches = defaultdict(set)
    for k, v in terms_dict.items():
        if any(i in source for i in v[0]):
            # Add that the `k` modality was found (e.g., 'mri').
            matches[k].add(None)
            # Scan `source` for subtypes (taking the opportunity to
            # look for those that may not be modality specific).
            if k in modality_subtypes:
                for v2 in modality_subtypes[k]:
                    if any(j in source for j in v2):
                        matches[k].add(cln(v2[0]))

    # 2. Look for subtype matches which can be used to infer the modality itself.
    for k, v in modality_specific_subtypes.items():
        for i in v:
            if any(j in source for j in i):
                matches[k].add(cln(i[0]))

    if len(matches.keys()) != 1:
        return None
    else:
        return {k: list(filter(None, v)) for k, v in matches.items()}


def _im_drop_contradictions(values):
    """

    Check for and eliminate contradictions.

    :param values:
    :return:
    """
    for c in contradictions:
        if all(i in values for i in c):
            values = [m for m in values if m not in c]
    return values


def _im_formatter(dictionary):
    """

    Format the output; form: 'formal modality name: subtype1; subtype2; ...'.

    :param dictionary:
    :return:
    """
    modality = list(dictionary.keys())[0]
    modality_formal_name = terms_dict[modality][-1]
    values = _im_drop_contradictions(set(dictionary[modality]))
    if not len(values):
        return modality_formal_name
    else:
        return "{0}: {1}".format(modality_formal_name, "; ".join(sorted(values)))


def _imaging_modality_guess(abstract, image_caption, image_mention):
    """

    Guess the imaging technology (modality) used to take the picture.
    Preference: ``image_caption`` > ``image_mention`` > ``abstract``.

    :param abstract: a text abstract.
    :type abstract: ``str``
    :param image_caption: an element from the 'image_caption' column.
    :type image_caption: ``str``
    :param image_mention: an element from the 'image_mention' column.
    :type image_mention: ``str``
    :return: imaging modality.
    :rtype: ``str`` or ``None``
    """
    # Try to return by scanning `image_caption`, `image_mention` and `abstract`
    for source in (image_caption, image_mention, abstract):
        if isinstance(source, str):
            scan_rslt = _im_scan(source=cln(source).lower())
            if isinstance(scan_rslt, dict) and len(scan_rslt.keys()) == 1:
                return _im_formatter(scan_rslt)
    else:
        return None  # capitulate


# ----------------------------------------------------------------------------------------------------------
# Image Plane
# ----------------------------------------------------------------------------------------------------------


def _image_plane_guess(image_caption):
    """

    Guess whether the plane of the image is 'axial', 'coronal' or 'sagittal'.

    :param image_caption: an element from the 'image_caption' column.
    :type image_caption: ``str``
    :return: see description.
    :rtype: ``str`` or ``None``
    """
    if not isinstance(image_caption, str):
        return None

    image_caption_clean = cln(image_caption).lower()
    plane_terms = [['axial', 'transverse'], ['coronal'], ['sagittal']]
    planes = [p[0] for p in plane_terms if any(i in image_caption_clean for i in p)]
    return planes[0] if len(planes) == 1 else None


# ----------------------------------------------------------------------------------------------------------
# Grids (inferred by the presence of enumerations in the text)
# ----------------------------------------------------------------------------------------------------------


def _extract_enumerations(input_str):
    """

    Extracts enumerations from strings.

    :param input_str: any string.
    :type input_str: ``str``
    :return: enumerations present in `input_str`. Returns ``None`` if ``input_str`` is empty.
    :rtype: ``None`` or ``list``

    :Example:

    >>> _extract_enumerations('(1a) here we see... 2. whereas here we see...')
    ...
    ['1a', '2']
    """
    # Block floating point numbers, e.g., '...there are peaks at 2.3 and 3.28.'
    # See: http://stackoverflow.com/a/4703409/4898004. (Dropped second * to ignore '2.', but still catch (2.3).
    floating_point_numbers_removed = cln(re.sub(r"[-+]?\d*\.\d+", " ", input_str))

    # Further clean the input
    cleaned_input = cln(floating_point_numbers_removed, extent=2).replace("-", "").lower()

    if not len(cleaned_input):
        return None

    # Define a list to populate
    enumerations = list()

    # Markers which denote the possible presence of an enumeration.
    markers_left = ("(", "[")
    markers_right = (".", ",", ")", "]")

    # Add a marker to the end of `cleaned_input`
    cleaned_input = cleaned_input + ")" if cleaned_input[-1] not in markers_right else cleaned_input

    # Candidate enumeration
    candidate = ''

    # Walk through the cleaned string
    for i in cleaned_input:
        if i in markers_left:
            candidate = ''
        elif i not in markers_right:
            candidate += i
        elif i in markers_right:
            if len(candidate) and len(candidate) <= 2 and all(i.isalnum() for i in candidate):
                if sum(1 for c in candidate if c.isdigit()) <= 2 and sum(1 for c in candidate if c.isalpha()) <= 3:
                    enumerations.append(candidate)
            candidate = ''

    return enumerations


def _enumerations_test(l):
    """

    Test if ``l`` is an enumeration.

    :param l: a list of possible enumerations, e.g., ``['1', '2', '3']`` (assumed to be sorted and all lower case).
    :type l: ``list``
    :return: whether or not ``l`` is an enumeration.
    :rtype: ``bool``
    """
    def alnum_check(char):
        """
        Check if `char` is plausibly part of an enumeration when
        it may be part of a 'mixed' sequence, e.g., `['1', '2', '2a']`."""
        if len(char) == 1 and char.isdigit() or char.isalpha():
            return True
        elif len(char) == 2 and sum(1 for i in char if i.isdigit()) == 1 and sum(1 for i in char if i.isalpha()) == 1:
            return True
        else:
            return False

    # Check for all 'i's, e.g., ['i', 'ii', 'iii', 'iv'] etc.
    if all(item in ('i', 'ii', 'iii') for item in l[:3]):
        return True
    # Check for letter enumeration
    elif list(l) == list(string.ascii_lowercase)[:len(l)]:
        return True
    # If the above check yielded a False, ``l`` must be junk, e.g., ``['a', 'ml', 'jp']``.
    elif all(i.isalpha() for i in l):
        return False
    # Check for numeric enumeration, e.g., ['1', '2', '3', '4'].
    elif order_set(l) == list(map(str, range(1, len(l) + 1))):
        return True
    # Block simple enumerations that do not start at 1, e.g., ['2', '3', '4'].
    elif all(i.isdigit() for i in l):
        return False
    # Check for a combination
    elif all(alnum_check(i) for i in l):
        return True
    else:
        return False


def _enumerations_guess(image_caption, enumerations_grid_threshold):
    """

    Look for grids based on the presence of enumerations in the image caption.

    :param image_caption:
    :type image_caption: ``str``
    :param enumerations_grid_threshold:
    :type enumerations_grid_threshold: ``int``
    :return: true if ``image_caption`` contains markers that indicate image enumeration.
    :rtype: ``bool``
    """
    # Clean the input
    image_caption_reduced_confusion = multi_replace(image_caption, ['i.e.,', 'i.e.', 'e.g.,', 'e.g.', 'e.x.,', 'e.x.'])
    image_caption_clean_lower = cln(image_caption_reduced_confusion, extent=2).lower()

    # Check for markers of enumeration like '(a-e)', '(a,e)', '(b and c)'.
    match_on = ["[(|\[][a-z]" + i + "[a-z][)|\]]" for i in ("-", "and")] + ["[a-z]~[a-z]"]
    for regex in match_on:
        if len(re.findall(regex, image_caption_clean_lower)):
            return True
    else:
        # Fall back to standard enumerations
        enums = _extract_enumerations(input_str=image_caption_reduced_confusion)
        if isinstance(enums, list) and _enumerations_test(enums) and len(enums) >= enumerations_grid_threshold:
            return True
        else:
            return False


# ----------------------------------------------------------------------------------------------------------
# Markers (inferred from the text)
# ----------------------------------------------------------------------------------------------------------


def _markers_guess(image_caption):
    """

    Currently detects the mention of 'arrows' and 'asterisks' in the image,
    based on the text.

    :param image_caption: an element from the 'image_caption' column.
    :type image_caption: ``str``
    :return: whether or not the image contains markers, based on the text.
    :rtype: ``bool``
    """
    features = set()
    image_caption_clean = cln(image_caption).lower()

    for term in ('arrow', 'asterisk'):
        # Look for arrows or asterisk.
        match_on = tuple(map(lambda x: x.format(term), (' {0} ', ' {0}', ' {0}s ', ' {0}s')))
        # Example: '...which are indicated by the asterisks...'
        if any(i in image_caption_clean for i in match_on):
            features.add("{0}s".format(term))
        # Example: '...along the left side (arrows)...'
        elif len(re.findall(r'[(|\[]' + term + 's?[)|\]|,|;]', image_caption_clean)):
            features.add("{0}s".format(term))
        # Example: '...to the left (red arrows)...'
        elif len(re.findall(r'[(|\[].*? ' + term + 's?[)|\]|,|;]', image_caption_clean)):
            features.add("{0}s".format(term))

    # More markers may be added in the future
    if 'asterisks' not in features:
        if any("({0})".format(i) in image_caption_clean for i in ['*']):
            features.add("asterisks")

    return list(features)


# ----------------------------------------------------------------------------------------------------------
# Image Problems (Detected by Analysing their Associated Text)
# ----------------------------------------------------------------------------------------------------------


def _problematic_image_features(image_caption, image_caption_unique, enumerations_grid_threshold=2):
    """

    Currently determines whether or not the image contains
    arrows, asterisks or grids (multiple images arranged as as a grid of images).

    :param image_caption: an element from the 'image_caption' column.
    :type image_caption: ``str``
    :param image_caption_unique: whether or not the caption is unique with respect to the 'uid'.
    :type image_caption_unique: ``bool``
    :param enumerations_grid_threshold: the number of enumerations required to 'believe' the image is actually a
                                        'grid' of images, e.g., '1a. here we. 1b, rather' would suffice if this
                                        parameter is equal to `2`.
    :type enumerations_grid_threshold: ``int``
    :return: information on the presence of problematic image features.
    :rtype: ``tuple``

    :Example:

    >>> _problematic_image_features('1. left (green asterisk). 2. right(arrow; red and blue)...', image_caption_unique=True)
    ...
    ('arrows', 'asterisks', 'grids')
    """
    # Note: if ``image_caption_unique`` is not true, then the caption likely do not refer to the
    # current image, but with multiple images for the given 'uid'.
    if not image_caption_unique or not isinstance(image_caption, str) or not len(cln(image_caption, extent=2)):
        return None

    features = []

    # Look for markers
    features += _markers_guess(image_caption)

    if _enumerations_guess(image_caption, enumerations_grid_threshold):
        features.append('grids')
    
    return tuple(sorted(features)) if len(features) else None


# ----------------------------------------------------------------------------------------------------------
# Exposed Feature Extract Tool
# ----------------------------------------------------------------------------------------------------------


def _background_extract(d):
    """

    Extract background information for a parsed abstract.

    :param d: the ``d`` dictionary as evolved inside ``feature_extract()``.
    :type d: ``dict``
    :return: the study background/history (or case information, e.g., 'case_report').
    :rtype: ``None`` or ``str``
    """
    # Extract the background/history
    if isinstance(d['parsed_abstract'], dict):
        background = d['parsed_abstract'].get('background', None)
        if background is None:
            # Try to use 'case' information instead.
            case_keys = [k for k in d['parsed_abstract'] if 'case' in k]
            if len(case_keys) == 1:
                background = d['parsed_abstract'][case_keys[0]]
    else:
        background = None

    return background


def feature_extract(x, list_of_diseases, image_caption_unique):
    """

    Tool to extract text features from patient summaries.

    See the description of ``OpeniInterface().pull()`` for additional information.

    :param x: series passed though Pandas' ``DataFrame().apply()`` method, e.g.,
              ``df.apply(lambda x: feature_extract(x, list_of_diseases), axis=1)``.

              .. note::

                   The dataframe must contain 'title', 'abstract', 'image_caption', 'image_mention'
                   and 'journal_title' columns.

    :type x: ``Pandas Series``
    :param list_of_diseases: a list of diseases (e.g., via ``DiseaseOntInterface().pull()['name'].tolist()``)
    :type list_of_diseases: ``list``
    :param image_caption_unique: whether or not the caption is unique with respect to the 'uid'.
    :type image_caption_unique: ``bool``
    :return: dictionary with the keys listed in the description.
    :rtype: ``dict``
    """
    # Note: it may be useful to use 'unescape' here, e.g., unescape(x[c])...but no reason to support doing so currently.
    # However, if this were to occur, 'abstract' would need to be excluded in the service of ``_abstract_parser()``.
    cols_to_unpack = ('abstract', 'problems', 'title', 'image_caption', 'image_mention', 'article_type')
    abstract, problems, title, image_caption, image_mention, article_type = [x[c] for c in cols_to_unpack]

    d = {'parsed_abstract': _abstract_parser(abstract)}
    background = _background_extract(d)

    # Extract diagnosis -- will typically only work for MedPix images
    d['diagnosis'] = d['parsed_abstract'].get('diagnosis', None) if isinstance(d['parsed_abstract'], dict) else None

    # Fall back
    if d['diagnosis'] is None:
        d['diagnosis'] = _disease_guess(problems=problems, title=title, background=background,
                                        # Block use of 'abstract' when the article is not clinical.
                                        abstract=abstract if article_type in CLINICAL_ARTICLE_TYPES else None,
                                        image_caption=image_caption if image_caption_unique else None,
                                        image_mention=image_mention,
                                        list_of_diseases=list_of_diseases)

    if isinstance(d['diagnosis'], str):
        d['diagnosis'] = d['diagnosis'].lower()

    pairs = [('age', _patient_age_guess), ('sex', _patient_sex_guess), ('illness_duration_years', _illness_duration_guess)]
    for (k, func) in pairs:
        d[k] = func(background, abstract, image_caption, image_mention)

    d['imaging_modality_from_text'] = _imaging_modality_guess(abstract, image_caption, image_mention)
    d['image_plane'] = _image_plane_guess(image_caption)
    d['image_problems_from_text'] = _problematic_image_features(image_caption, image_caption_unique)
    d['ethnicity'], eth_sex = _ethnicity_guess(background, abstract, image_caption, image_mention)

    if d['sex'] is None and eth_sex is not None:
        d['sex'] = eth_sex

    return d



"""


"""
import numpy as np
from numpy import array
from PIL import Image
from scipy.misc import imresize
from biovida.images import OpeniInterface
from biovida.images import OpeniImageProcessing

from sklearn.metrics import matthews_corrcoef

path = "/Users/tariq/Desktop/test_image.png"


# Model
opi = OpeniInterface()
ip = OpeniImageProcessing(opi, db_to_extract='cache_records_db')
model = ip._ircnn.model
image_shape = ip._ircnn.image_shape

# Test
image = array(Image.open(path))


def image_reformat(image, image_shape=(224, 224)):
    image = imresize(image, image_shape)
    image = image.transpose((2,0,1))
    image = image.astype('float32')
    image = image / 255
    return np.expand_dims(image, axis=0)

x_test = array([image_reformat(image)]*3)
y_test = np.array([1, 0])

out = array(model.predict_proba(x_test[0]))


acc = []
accuracies = []
threshold = np.arange(0.1, 0.9, 0.1)
best_threshold = np.zeros(out.shape[1])
for i in range(out.shape[1]):
    y_prob = np.array(out[:, i])
    for j in threshold:
        y_pred = [1 if prob >= j else 0 for prob in y_prob]
        acc.append(matthews_corrcoef(y_test[:, i], y_pred))
    acc = np.array(acc)
    index = np.where(acc == acc.max())
    accuracies.append(acc.max())
    best_threshold[i] = threshold[index[0][0]]
    acc = []



test_data = "/Users/tariq/Desktop/dataset"

to_location





































# Expose key Tools at the Subpackage level
from biovida.images.image_cache_mgmt import image_divvy
from biovida.images.image_cache_mgmt import image_delete

# Expose key Class at the Subpackage level
from biovida.images.openi_interface import OpeniInterface
from biovida.images.image_processing import OpeniImageProcessing
from biovida.images.cancer_image_interface import CancerImageInterface
from biovida.images.models.image_classification import ImageClassificationCNN



# coding: utf-8

"""

    Tools to Clean Raw Open-i Text
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
import numpy as np
import pandas as pd
from bs4 import BeautifulSoup
from collections import Counter

# General Support tools
from biovida.support_tools.support_tools import cln, unescape, remove_html_bullet_points

# Image Support Tools
from biovida.images._image_tools import NoResultsFound

# General Support Tools
from biovida.support_tools.support_tools import tqdm, camel_to_snake_case, reverse_comma, order_set

# Open-i API Parameters Information
from biovida.images._interface_support.openi.openi_parameters import (openi_image_type_params,
                                                                      openi_article_type_params,
                                                                      openi_image_type_modality_full)

from biovida.images._interface_support.openi._openi_image_id_processing import image_id_short_gen
from biovida.images._interface_support.openi._openi_text_feature_extraction import (CLINICAL_ARTICLE_TYPES,
                                                                                    feature_extract)

# Other BioVida APIs
from biovida.diagnostics.disease_ont_interface import DiseaseOntInterface


# ----------------------------------------------------------------------------------------------------------
# Abstract Cleaning
# ----------------------------------------------------------------------------------------------------------


def abstract_cleaner(abstract):
    """

    Clean ``abstract`` by converting the HTML into standard text.

    :param abstract: an abstract obtained via. the Open-i API.
    :type abstract: ``str``
    :return: cleaned ``abstract``
    :rtype: ``str``
    """
    if not isinstance(abstract, str):
        return np.NaN

    soup = BeautifulSoup(remove_html_bullet_points(abstract).replace("<b>", ". "), 'lxml')
    cleaned = soup.text.replace(" ; ", " ").replace("..", ".").replace(".;", ";").strip()
    return cleaned.strip(".") + "."


# ----------------------------------------------------------------------------------------------------------
# Problems Cleaning
# ----------------------------------------------------------------------------------------------------------


def problems_cleaner(problems):
    """
    
    Clean strings in the 'problems'
    column by:
    
    - removing duplicates
    - reversing items with commas
    
    :param problems: an item from the 'problems' column
    :return: cleaned ``problems``
    :rtype: ``str``
    
    :Example:
    
    >>> problems_cleaner("Lung; Fractures, Bone")
    ...
    'lung; bone fractures'
    
    """
    if not isinstance(problems, str):
        return problems

    cleaned = order_set(problems.lower().split(";"))
    return "; ".join([reverse_comma(s=cln(i)) for i in cleaned])


# ----------------------------------------------------------------------------------------------------------
# Article Type Lookup
# ----------------------------------------------------------------------------------------------------------


def article_type_lookup(article_type_abbrev):
    """

    Lookup the full article type from the shorthand used by the Open-i API.

    :param article_type_abbrev: values as passed via. ``data_frame['article_type'].map(...)``.
    :return: the full article type name.
    :rtype: ``str``
    """
    # Note: It appears that, at the time of writing (March 16, 2017), Open-i has switched
    # to transforming this data on their end. However, this function remains as a safeguard.
    if not isinstance(article_type_abbrev, str):
        return article_type_abbrev

    return openi_article_type_params.get(cln(article_type_abbrev).lower(), article_type_abbrev)


# ----------------------------------------------------------------------------------------------------------
# Handle Missing Columns
# ----------------------------------------------------------------------------------------------------------


def _df_add_missing_columns(data_frame):
    """

    Add 'license_type', 'license_url' and 'image_caption_concepts' columns
    if they do not exist in ``data_frame``

    :param data_frame: the dataframe evolved inside ``openi_raw_extract_and_clean``.
    :type data_frame: ``Pandas DataFrame``
    :return: see description.
    :rtype: ``Pandas DataFrame``
    """
    # Handle cases where some searches (e.g., collection='pubmed')
    # Open-i does not return these columns (not clear why...).
    to_check = ['license_type', 'image_caption_concepts', 'pub_med_url', 'license_url',
                'impression', 'note', 'medpix_article_id', 'medpix_figure_id', 'medpix_image_url']

    for c in to_check:
        if c not in data_frame.columns:
            data_frame[c] = [np.NaN] * data_frame.shape[0]
    return data_frame


# ----------------------------------------------------------------------------------------------------------
# Tool to Limit to Clinical Cases
# ----------------------------------------------------------------------------------------------------------


def _apply_clinical_case_only(data_frame):
    """

    Remove records (dataframe rows) which are not of clinical encounters.

    Note: this is here, and not simply part of defining a search with ``openi_interface()._OpeniImages().search()``
    because Open-i API's 'article_type' (&at) parameter does not have an 'encounter' option.

    :param data_frame: the ``data_frame`` as evolved in ``openi_raw_extract_and_clean()``.
    :type data_frame: ``Pandas DataFrame``
    :return: see description.
    :rtype: ``Pandas DataFrame``
    """
    def test(article_type):
        if isinstance(article_type, str) and article_type in CLINICAL_ARTICLE_TYPES:
            return True
        else:
            return False

    data_frame = data_frame[data_frame['article_type'].map(test)].reset_index(drop=True)

    if data_frame.shape[0] == 0:
        raise NoResultsFound("\nNo results remained after the `clinical_cases_only=True` restriction was applied.\n"
                             "Consider setting `pull()`'s `clinical_cases_only` parameter to `False`.")
    return data_frame


# ----------------------------------------------------------------------------------------------------------
# Make Hashable
# ----------------------------------------------------------------------------------------------------------


def _iterable_cleaner(iterable):
    """

    Clean terms in an iterable and return as a tuple.

    :param iterable: a list of terms.
    :type iterable: ``tuple`` or ``list``
    :return: a cleaned tuple of strings.
    :rtype: ``tuple`` or ``type(iterable)``
    """
    if isinstance(iterable, (list, tuple)):
        return tuple([cln(unescape(i)) if isinstance(i, str) else i for i in iterable])
    else:
        return iterable


def _df_make_hashable(data_frame):
    """

    Ensure the records dataframe can be hashed (i.e., ensure ``pandas.DataFrame.drop_duplicates`` does not fail).

    :param data_frame: the dataframe evolved inside ``openi_raw_extract_and_clean``.
    :type data_frame: ``Pandas DataFrame``
    :return: ``data_frame`` corrected such that all columns considered can be hashed.
    :rtype: ``Pandas DataFrame``
    """
    # Escape HTML elements in the 'image_caption' and 'image_mention' columns.
    for c in ('image_caption', 'image_mention'):
        data_frame[c] = data_frame[c].map(lambda x: cln(unescape(x)) if isinstance(x, str) else x, na_action='ignore')

    # Clean mesh terms
    for c in ('mesh_major', 'mesh_minor', 'image_caption_concepts'):
        data_frame[c] = data_frame[c].map(_iterable_cleaner, na_action='ignore')

    # Convert all elements in the 'license_type' and 'license_url' columns to string.
    for c in ('license_type', 'license_url'):
        data_frame[c] = data_frame[c].map(lambda x: "; ".join(map(str, x)) if isinstance(x, (list, tuple)) else x,
                                          na_action='ignore')

    return data_frame


# ----------------------------------------------------------------------------------------------------------
# Cleaning
# ----------------------------------------------------------------------------------------------------------


def _data_frame_fill_nan(data_frame):
    """

    Replace terms that are synonymous with NA with NaNs.

    :param data_frame: the dataframe evolved inside ``openi_raw_extract_and_clean``.
    :type data_frame: ``Pandas DataFrame``
    :return: ``data_frame`` with terms that are synonymous with NA replaced with NaNs.
    :rtype:  ``Pandas DataFrame``
    """
    to_nan = ('[nN]ot [aA]vailable.?', '[Nn]one.?', '[Nn]/[Aa]', '[Nn][Aa]',
              '[Ii][Nn] [Pp][Rr][Oo][Gg][Rr][Ee][Ss][Ss].?')
    # Anchor (i.e., exact matches are required, e.g., "NA" --> NaN, but "here we see NA" will not be converted).
    anchored_to_nan = map(lambda x: "^{0}$".format(x), to_nan)
    data_frame = data_frame.replace(dict.fromkeys(anchored_to_nan, np.NaN), regex=True)

    # Replace the 'replace this - ' placeholder with NaN
    data_frame['image_caption'] = data_frame['image_caption'].map(
        lambda x: np.NaN if isinstance(x, str) and cln(x).lower().startswith('replace this - ') else x,
        na_action='ignore')

    return data_frame.fillna(np.NaN)


def _data_frame_clean(data_frame, verbose):
    """

    Clean the text information.

    :param data_frame: the dataframe evolved inside ``openi_raw_extract_and_clean``.
    :type data_frame: ``Pandas DataFrame``
    :param verbose: if ``True`` print additional details.
    :type verbose: ``bool``
    :return: cleaned ``data_frame``.
    :rtype:  ``Pandas DataFrame``
    """
    # Clean the abstract
    data_frame['abstract'] = data_frame['abstract'].map(abstract_cleaner)

    # Clean the 'image_caption'
    data_frame['image_caption'] = data_frame['image_caption'].map(
        lambda x: cln(BeautifulSoup(x, 'lxml').text) if isinstance(x, str) else x, na_action='ignore')

    # Add the full name for modalities (before the 'image_modality_major' values are altered below).
    data_frame['modality_full'] = data_frame['image_modality_major'].map(
        lambda x: openi_image_type_modality_full.get(cln(x).lower(), x), na_action='ignore')

    # Make the type of Imaging technology type human-readable.
    data_frame['image_modality_major'] = data_frame['image_modality_major'].map(
        lambda x: openi_image_type_params.get(cln(x).lower(), x), na_action='ignore')

    # Label the number of instance of repeating 'uid's.
    data_frame = image_id_short_gen(data_frame)

    # Replace missing Values with with NaN and Return
    return _data_frame_fill_nan(data_frame)


# ----------------------------------------------------------------------------------------------------------
# Image Caption Count Analysis
# ----------------------------------------------------------------------------------------------------------


def _unique_image_caption_dict_gen(data_frame, verbose):
    """

    Generate a dictionary of the form:
    ``{'uid': {'image caption goes here': unique (bool), ...}, ...}``

    :param data_frame: as evolved in ``openi_raw_extract_and_clean()``
    :type data_frame: ``Pandas DataFrame``
    :param verbose: if ``True`` print additional details.
    :type verbose: ``bool``
    :return: see description
    :rtype: ``dict``
    """
    large_data_frame = 25000  # after this number of rows, processing time is notable.

    def counter_wrapper(l):
        l_cleaned = filter(lambda x: isinstance(x, str) and len(cln(x)), l)
        count_dict = dict(Counter(l_cleaned))
        return {k: v == 1 for k, v in count_dict.items()} if count_dict else None

    def counter_wrapper_apply(row):
        return counter_wrapper(row['image_caption'].tolist())

    if len(data_frame) >= large_data_frame and verbose:
        print("\nAnalyzing Image Caption Frequency...")

    d = data_frame.groupby('uid').apply(counter_wrapper_apply).to_dict()

    return {k: v for k, v in d.items() if v is not None}


# ----------------------------------------------------------------------------------------------------------
# Outward Facing Tool
# ----------------------------------------------------------------------------------------------------------


def openi_raw_extract_and_clean(data_frame, clinical_cases_only, verbose, cache_path):
    """

    Extract features from, and clean text of, ``data_frame``.

    :param data_frame: the dataframe evolved inside ``biovida.images.openi_interface._OpeniRecords().records_pull()``.
    :rtype data_frame: ``Pandas DataFrame``
    :param clinical_cases_only: if ``True`` require that the data harvested is of a clinical case. Specifically,
                                this parameter requires that 'article_type' is one of: 'encounter', 'case_report'.
                                Defaults to ``True``.
    :type clinical_cases_only: ``bool``
    :param verbose: if ``True`` print additional details.
    :type verbose: ``bool``
    :param cache_path: path to the location of the BioVida cache. If a cache does not exist in this location,
                       one will created. Default to ``None``, which will generate a cache in the home folder.
    :type cache_path: ``str``
    :return: see description.
    :rtype:  ``Pandas DataFrame``
    """
    data_frame.columns = list(map(lambda x: camel_to_snake_case(x).replace("me_sh", "mesh"), data_frame.columns))
    data_frame = _df_add_missing_columns(data_frame)

    data_frame['article_type'] = data_frame['article_type'].map(article_type_lookup, na_action='ignore')
    data_frame['problems'] = data_frame['problems'].map(problems_cleaner, na_action='ignore')

    if clinical_cases_only:
        data_frame = _apply_clinical_case_only(data_frame)

    # Ensure the dataframe can be hashed (i.e., ensure pandas.DataFrame.drop_duplicates does not fail).
    data_frame = _df_make_hashable(data_frame)

    list_of_diseases = DiseaseOntInterface(cache_path=cache_path, verbose=verbose).pull()['name'].tolist()
    unique_image_caption_dict = _unique_image_caption_dict_gen(data_frame=data_frame, verbose=verbose)

    def feature_extract_wrapper(row):
        caption_unique_bool = unique_image_caption_dict.get(row['uid'], {}).get(row['image_caption'], False)
        return feature_extract(row, list_of_diseases=list_of_diseases, image_caption_unique=caption_unique_bool)

    # Run Feature Extracting Tool and Join with `data_frame`.
    extract = [feature_extract_wrapper(row) for _, row in tqdm(data_frame.iterrows(), desc='Processing Records',
                                                               total=len(data_frame), disable=not verbose)]
    extract_df = pd.DataFrame(extract)

    if not any(c in data_frame.columns for c in extract_df.columns):
        data_frame = data_frame.join(extract_df, how='left')
    else:
        for c in extract_df.columns:
            data_frame[c] = extract_df[c]

    return _data_frame_clean(data_frame, verbose=verbose)



# coding: utf-8

"""

    Convert DICOM Data into a Python Dictionary
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
# Also see: https://github.com/darcymason/pydicom/issues/319.
import os
from biovida.support_tools.support_tools import cln, dicom


def _extract_numeric(value):
    """

    :param value:
    :return:
    """
    return float("".join((i for i in value if i.isdigit() or i == '.')))


def parse_age(value):
    """

    :param value:
    :return:
    """
    if not isinstance(value, str):
        raise TypeError('`value` must be a string.')
    elif len(value) > 4:
        return value

    if 'y' in value.lower():
        return _extract_numeric(value)
    elif 'm' in value.lower():
        return _extract_numeric(value) / 12.0
    else:
        return value


def parse_string_to_tuple(value):
    """

    :param value:
    :type value: ``str``
    :return:
    """
    braces = [['[', ']'], ['(', ')']]
    for (left, right) in braces:
        if left in value and right in value:
            value_split = value.replace(left, "").replace(right, "").split(",")
            value_split_cln = list(filter(None, map(cln, value_split)))
            if len(value_split_cln) == 0:
                return None
            try:
                to_return = tuple(map(_extract_numeric, value_split_cln))
            except:
                to_return = tuple(value_split_cln)
            return to_return[0] if len(to_return) == 1 else to_return
    else:
        raise ValueError("Cannot convert `value` to a tuple.")


def dicom_value_parse(key, value):
    """
    
    Try to convert ``value`` to a numeric or tuple of numerics.

    :param key:
    :param value:
    :return:
    """
    value = cln(str(value).replace("\'", "").replace("\"", ""))

    if not len(value) or value.lower() == 'none':
        return None

    if key.lower().endswith(' age') or key == 'PatientAge':
        try:
            return parse_age(value)
        except:
            return value
    else:
        try:
            return int(value)
        except:
            try:
                return float(value)
            except:
                try:
                    return parse_string_to_tuple(value)
                except:
                    return value


def dicom_object_dict_gen(dicom_object):
    """

    :param dicom_object:
    :type dicom_object: ``dicom.FileDataset``
    :return:
    """
    d = dict()
    for k in dicom_object.__dir__():
        if not k.startswith("__") and k != 'PixelData':
            try:
                value = dicom_object.data_element(k).value
                if type(value).__name__ != 'Sequence':
                    d[k] = dicom_value_parse(key=k, value=value)
            except:
                pass
    return d


def dicom_to_dict(dicom_file):
    """

    Convert the metadata associated with ``dicom_file`` into a python dictionary

    :param dicom_file: a path to a dicom file or the yield of ``dicom.read_file(FILE_PATH)``.
    :type dicom_file: ``FileDataset`` or ``str``
    :return: a dictionary with the dicom meta data.
    :rtype: ``dict``
    """
    if isinstance(dicom_file, str):
        if not os.path.isfile(dicom_file):
            raise FileNotFoundError("Could not locate '{0}'.".format(dicom_file))
        dicom_object = dicom.read_file(dicom_file)
    elif type(dicom_file).__name__ == 'FileDataset':
        dicom_object = dicom_file
    else:
        raise TypeError("`dicom_file` must be of type `dicom.FileDataset` or a string.")

    return dicom_object_dict_gen(dicom_object)



import os
from setuptools import setup, find_packages


def read(fname):
    try:
        return open(os.path.join(os.path.dirname(__file__), fname)).read()
    except:
        return 'Please see: https://github.com/TariqAHassan/BioVida.'


setup(
    name='biovida',
    version='0.1.1',
    author='Tariq A. Hassan',
    author_email='laterallattice@gmail.com',
    description=('Automated BioMedical Information Curation for Machine Learning Applications.'),
    long_description=read('docs/README.rst'),
    license='BSD',
    keywords='machine-learning, biomedical-informatics, data-science, bioinformatics, imaging-informatics',
    url='https://github.com/TariqAHassan/BioVida.git',
    download_url='https://github.com/TariqAHassan/BioVida/archive/v0.1.1.tar.gz',
    packages=find_packages(exclude=("data.*",
                                    "data",
                                    "tests.*",
                                    "tests")
                           ),
    install_requires=['bs4', 'h5py', 'keras', 'lxml', 'numpy', 'pandas', 'Pillow',
                      'pydicom', 'requests', 'scikit-image', 'scipy', 'theano', 'tqdm'],
    classifiers=['Development Status :: 3 - Alpha',
                 'Natural Language :: English',
                 'Intended Audience :: Science/Research',
                 'Programming Language :: Python :: 3.4',
                 'Programming Language :: Python :: 3.5',
                 'Programming Language :: Python :: 3.6',
                 'License :: OSI Approved :: BSD License'
    ],
    include_package_data=True
)



add_to_path = "/Users/tariq/Google Drive/Programming Projects/BioVida"
training_data_path = "/Users/tariq/Desktop/biovida_training/actual_training_images"
biovida_path = "/Users/tariq/Google Drive/Programming Projects/BioVida"


# coding: utf-8

"""

    BioVida Cache Management Tools
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
# Note: this module contains the tools required to construct the caches needed by BioVida.
import os
import requests
from PIL import Image

# General Support Tools
from biovida.support_tools.support_tools import combine_dicts, list_to_bulletpoints


def _medpix_logo_download(save_path,
                          verbose,
                          image_web_address='https://openi.nlm.nih.gov/imgs/512/341/1544/MPX1544_synpic54565.png',):
    """

    Download and save the MedPix logo from a representative image.
    (The image is needed for image processing).

    :param image_web_address: the location of the image which contains the MedPix Logo
    :type image_web_address: ``str``
    :param verbose: notify user of directory that have been created. Defaults to True.
    :type verbose: ``bool``
    :return: a dictionary of the form {'medpix_logo': PATH TO THE LOGO}.
    :rtype: ``dict``
    """
    full_save_path = os.path.join(save_path, "medpix_logo.png")

    if not os.path.isfile(full_save_path):
        # Get representative images
        image = Image.open(requests.get(image_web_address, stream=True).raw)
        # Crop and Save
        image_cropped = image.crop((406, 6, 502, 27))
        image_cropped.save(full_save_path)
        if verbose:
            print("\nThe MedPix Logo, required for processing images from Open-i, "
                  "has been downloaded to:\n\n{0}\n".format(list_to_bulletpoints([full_save_path])))

    return {'medpix_logo': full_save_path}


def _sub_directory_creator(root_path, to_create):
    """

    :param root_path:
    :param to_create:
    :return:
    """
    created_dirs = dict()

    # Create sub directories
    for sub_dir in to_create:
        # Check if the directory exists
        if not os.path.isdir(os.path.join(root_path, sub_dir)):
            os.makedirs(os.path.join(root_path, sub_dir))
            # Record sub_dir's full path
            created_dirs[(sub_dir, True)] = (os.sep).join(os.path.join(root_path, sub_dir).split(os.sep)[-2:])
        else:
            # Note that it was not created
            created_dirs[(sub_dir, False)] = (os.sep).join(os.path.join(root_path, sub_dir).split(os.sep)[-2:])

    return created_dirs


def _created_notice(created_list, system_path):
    """

    :param created_list:
    :param system_path:
    :return:
    """
    if len(created_list):
        print("The following directories were created:\n%s\nin: '%s'." % \
              ("".join(["  - " + i + "\n" for i in created_list]), system_path + os.sep))
        print("\n")


def _directory_creator(cache_path=None, verbose=True):
    """

    Tool to create directories needed by BioVida.

    Required:
      - biovida_cache
      - biovida_cache/images_cache
      - biovida_cache/genomics_cache
      - biovida_cache/diagnostics_cache

    :param cache_path: path to create to create the `BioVida` cache.
                       If ``None``, the home directory will be used. Defaults to None.
    :type cache_path: ``str`` or ``None``
    :param verbose: notify user of directory that have been created. Defaults to True.
    :type verbose: ``bool``
    :return: the root path (i.e., path the cache itself).
    :rtype: ``str``
    """
    # Record of dirs created.
    created_dirs = list()

    # Clean `cache_path`
    if cache_path is not None and cache_path is not None:
        cache_path_clean = (cache_path.strip() if not cache_path.endswith(os.sep) else cache_path.strip()[:-1])
    else:
        cache_path_clean = cache_path

    # Set the base path to the home directory if `cache_path_clean` does not exist
    if isinstance(cache_path_clean, str) and not os.path.isdir(cache_path_clean):
        raise FileNotFoundError("[Errno 2] No such file or directory: '{0}'.".format(cache_path_clean))
    elif not (isinstance(cache_path_clean, str) and os.path.isdir(cache_path_clean)):
        base_path = os.path.expanduser("~")
    else:
        base_path = cache_path_clean

    # Set the 'biovida_cache' directory to be located in the home folder
    root_path = os.path.join(base_path, "biovida_cache")
    # If needed, create
    if not os.path.isdir(root_path):
        # Create main cache folder
        os.makedirs(os.path.join(base_path, "biovida_cache"))
        # Note its creation
        created_dirs.append("biovida_cache")

    # Check if 'search', 'images', 'genomics' and 'diagnostics' caches exist, if not create them.
    sub_dirs_made = _sub_directory_creator(root_path, ['images_cache', 'genomics_cache', 'diagnostics_cache'])

    # Record Created Dirs
    created_dirs += {k: v for k, v in sub_dirs_made.items() if k[1] is True}.values()

    # Print results, if verbose is True
    if verbose and len(created_dirs):
        _created_notice(created_dirs, base_path)

    return root_path


def _add_to_create_nest(nest, record_dict, verbose):
    """

    Add `nest` directories, i.e., create directories within directories delineated in `to_create` (below)..

    :param nest:
    :param record_dict:
    :return:
    :rtype: ``dict``
    """
    created = list()
    if isinstance(nest, (list, tuple)):
        for (to_create_dir, new_nested_dir) in nest:
            new_dir_name = os.path.join(record_dict[to_create_dir], new_nested_dir)
            if not os.path.exists(new_dir_name):
                os.makedirs(new_dir_name)
                created.append(new_dir_name)
            record_dict[new_nested_dir] = new_dir_name

    if verbose and len(created):
        print("The following nested directories were also created:\n{0}\n".format(list_to_bulletpoints(created)))

    return record_dict


def package_cache_creator(sub_dir, to_create, cache_path=None, nest=None, verbose=True, requires_medpix_logo=False):
    """

    Create a cache for a given ``sub_dir``. If no biovida path exists in ``cache_path``,
    all of the following caches will be created: 'search', 'images', 'genomics' and 'diagnostics'.

    :param sub_dir: e.g., 'images' (do not include "_cache").
                    Must be one of: 'search', 'images', 'genomics', 'diagnostics'.
    :type sub_dir: ``str``
    :param to_create: subdirectories within ``sub_dir`` to be created.
    :type to_create: ``iterable``
    :param cache_path: local path to the desired location of the cache.
                       If ``None``, will default to the the home directory. Defaults to ``None``.
    :type cache_path: ``str`` or ``None``
    :param nest: a list of the form [(to_create item, new nested directory name), ...]
    :type nest: ``list of iterables`` or ``None``
    :param verbose: If True, print updates. Defaults to True.
    :type verbose: ``bool``
    :param requires_medpix_logo: if ``True``, download the medpix logo.
    :type requires_medpix_logo: ``bool``
    :return: tuple of the form ``(local path to `sub_dir`, `record_dict`)``,
             where ``record_dict`` is of the form ``{to_create[0]: 'PATH_1', to_create[1]: 'PATH_2', ...}``.
             Note: ``record_dict`` if ``requires_medpix_logo`` is ``True``, a 'medpix_logo' key will also be present.
    :rtype: ``tuple``
    """
    # Check `sub_dir` is an allowed type
    allowed_sub_dirs = ('search', 'images', 'genomics', 'diagnostics')
    if sub_dir not in allowed_sub_dirs:
        raise ValueError("`sub_dir` must be one of:\n{0}".format(list_to_bulletpoints(allowed_sub_dirs)))

    # Check `to_create` is a ``list`` or ``tuple`` with nonzero length
    if not isinstance(to_create, (list, tuple)) or not len(to_create):
        raise AttributeError("`to_create` must be a `list` or `tuple` with a nonzero length.")

    # Create main
    root_path = _directory_creator(cache_path, verbose)

    # The full path to the subdirectory
    sub_dir_full_path = os.path.join(root_path, sub_dir.replace("/", "").strip() + "_cache")

    # Ask for sub directories to be created
    package_created_dirs = _sub_directory_creator(sub_dir_full_path, to_create)

    # New dirs created
    new = {k: v for k, v in package_created_dirs.items() if k[1] is True}

    # Print record of files created, if verbose is True
    if verbose and len(new.values()):
        _created_notice(new.values(), root_path)

    # Render a hash map of `cache_path` - to - local address
    record_dict = {k[0]: os.path.join(sub_dir_full_path, v.split(os.sep)[-1]) for k, v in package_created_dirs.items()}
    record_dict['ROOT_PATH'] = root_path

    # Add nested directories, if any
    record_dict_nest = _add_to_create_nest(nest, record_dict, verbose)
    
    # Download the medpix logo to `sub_dir`, if `sub_dir` is 'images'.
    if sub_dir == 'images' and requires_medpix_logo:
        if 'aux' not in record_dict_nest.keys():
            raise ValueError("`nest` (or `to_create`) must contain 'aux' if `requires_medpix_logo` is `True`.")
        medpix_logo_location = _medpix_logo_download(record_dict_nest['aux'], verbose=verbose)
        # Add the path to the logo to `record_dict_nest`
        record_dict_nest = combine_dicts(record_dict_nest, medpix_logo_location)

    # Return full path & the above mapping
    return sub_dir_full_path, record_dict_nest



# coding: utf-8

"""

    General Support Tools for Open-i Data Processing
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
import re
from itertools import chain
from six.moves.urllib.parse import urlsplit  # handle Python 2

# General Support tools
from biovida.support_tools.support_tools import cln, items_null
from biovida.support_tools._support_data import ages_as_tuples_rev

# Regex for ``extract_float()``
non_decimal = re.compile(r'[^\d.]+')

ages_as_tuples_rev_spaced = [(" {0} ".format(w), str(n)) for (w, n) in ages_as_tuples_rev]


# ----------------------------------------------------------------------------------------------------------
# Data
# ----------------------------------------------------------------------------------------------------------


nonessential_openi_columns = [
    # Columns to drop when returning
    # '..._short' dataframes attributes
    # of the OpeniInterface class.
    'biovida_version',
    'detailed_query_url',
    'fulltext_html_url',
    'get_article_figures',
    'image_id',
    'img_grid150',
    'img_large',
    'img_thumb',
    'img_thumb_large',
    'pmc_url',
    'pub_med_url',
    'license_url',
    'medpix_article_id',
    'medpix_figure_id',
    'medpix_image_url',
    'similar_in_collection',
    'similar_in_results'
]


possible_openi_image_processing_cols = [
    # Columns that can be generated by
    # the ImageProcessing Class.
    'grayscale',
    'medpix_logo_bounding_box',
    'hbar',
    'hborder',
    'vborder',
    'upper_crop',
    'lower_crop',
    'visual_image_problems',
    'valid_image'
]


grayscale_openi_modalities = [
    # Modalities, as enumerated in
    # the 'image_modality_major'
    # column which are grayscale.
    # MRI is assumed to be structural.
    'ct',
    'mri',
    'ultrasound',
    'x_ray',
]


# ----------------------------------------------------------------------------------------------------------
# Tools
# ----------------------------------------------------------------------------------------------------------


class ImageProblemBasedOnText(Exception):
    pass


def iter_join(t, join_on="_"):
    """

    Convert an iterabe (``t``) to a string by joining on ``join_on``.

    :param t: any iterable.
    :type t: ``iterable``
    :param join_on: a string to join ``t`` on.
    :type join_on: ``str``
    :return: ``t`` as a string.
    :rtype: ``str``
    """
    return join_on.join(t) if isinstance(t, (list, tuple)) else t


def null_convert(i):
    """

    Convert ``i`` to ``None`` if it is void else ``i`` is returned.

    :param i: any.
    :type i: ``any``
    :return: None ``i`` is void (e.g., '' or []) else ``i``.
    :rtype: ``None`` else ``type(i)``
    """
    return None if not i else i


def url_combine(url1, url2):
    """

    Combines two urls into one.

    :param url1: a URL.
    :type url1: ``str``
    :param url2: a URL.
    :type url2: ``str``
    :return: a string of the form ``url1/url2``.
    :rtype: ``None`` or ``str``
    """
    if any(x is None or items_null(x) for x in [url1, url2]):
        return None
    else:
        left = url1[:-1] if url1.endswith("/") else url1
        right = url2[1:] if url2.startswith("/") else url2
        return "{0}/{1}".format(left, right)


def item_extract(i, list_len=1):
    """

    Extract the first item in a list or tuple, else return ``None``.

    :param i: any list or tuple.
    :type i: ``list`` or ``tuple``
    :param list_len: required length of ``i``. Defaults to `1`.
    :type list_len: ``int``
    :return: the first element in ``i`` if ``i`` is a ``list`` or ``tuple`` and ``len(i) == list_len``.
    :rtype: ``type(i[0])`` or ``None``
    """
    return i[0] if isinstance(i, (list, tuple)) and len(i) == list_len else None


def extract_float(i):
    """

    Extract a d from a string ``i``

    :param i: a string.
    :type i: ``str``
    :return: the floating point number in ``i`` as a string.
    :rtype: ``str``
    """
    # Source: http://stackoverflow.com/a/947789/4898004
    return non_decimal.sub('', i)


def multiple_decimal_remove(s):
    """

    Remove multiple decimal from a string (``s``)

    :param s: a string.
    :type s: ``str``
    :return: ``s`` with only one.
    :rtype: ``None`` or ``str``
    """
    if isinstance(s, (int, float)):
        return s

    s_cln = cln(extract_float(str(s)), extent=2)
    if s_cln.startswith('.') and s_cln.endswith('.'):
        return None

    s_repeats_rmv = re.sub(r'\.+', '.', s_cln)
    if s_repeats_rmv.count('.') > 1:
        return None
    else:
        return cln(s_repeats_rmv, extent=2)


def filter_unnest(l, filter_for=None):
    """

    Filters out ``filter_for`` and flattens ``l``.

    :param l: a 2D iterable.
    :type l: ``iterable``
    :param filter_for: items to be removed. Defaults to ``None``.
    :type filter_for: any
    :return: ``l`` flattened with ``filter_for`` removed.
    :rtype: ``list``
    """
    return list(chain(*filter(filter_for, l)))


def num_word_to_int(input_str):
    """

    Replace natural numbers from 1 to 105 with integers.

    :param input_str: any string.
    :rtype input_str: ``str``
    :return: ``input_str`` with all numbers from 1-130 in natural language (e.g., 'twenty-five')
             with the integer equivalent.
    :rtype: ``str``

    :Example:

    >>> num_word_to_int(input_str="the disease was present from twenty  five until their death at eighty-seven.")
    ...
    'the disease was present from 25 until their death at 87'
    ...
    # However, the simple presence of the substring 'one' doesn't confuse the procedure.
    >>> num_word_to_int(input_str="the oneway (sic) street is closed due to construction.")
    ...
    'the oneway (sic) street is closed due to construction.'
    ...
    # It can also handle common punctuation (though it won't conserve it -- not currently needed however).
    >>> num_word_to_int(input_str="The ring was the only one, in all of history, to xyz")
    ...
    'The ring was the only 1 in all of history, to xyz'
    ...
    # It is also not confused by common expressions (assuming the number is < 20).
    >>> num_word_to_int(input_str="Robust one-shot learning is one of the ultimate goal of AI research.")
    'Robust one-shot learning is 1 of the ultimate goal of AI research.'
    """
    # Note: this is rather crude, but far faster than the prior regex solution.

    input_str_clean = " ".join(input_str.split()).strip().strip(".").strip(";")
    input_str_formatted = " {0} ".format(input_str_clean)

    for (w, n) in ages_as_tuples_rev_spaced:
        formmated_n = " {0} ".format(n)
        input_str_formatted = input_str_formatted.replace(w, formmated_n)
        input_str_formatted = input_str_formatted.replace(w.replace("-", " "), formmated_n)

        formmated_n = " {0}".format(n)
        input_str_formatted = input_str_formatted.replace("{0},".format(w.rstrip()), formmated_n)
        input_str_formatted = input_str_formatted.replace("{0},".format(w.replace("-", " ").rstrip()), formmated_n)
    return " ".join(input_str_formatted.split())


def url_path_extract(url):
    """

    Extracts the path for a given URL.

    :param url: a Uniform Resource Locator (URL).
    :rtype url: ``str``
    :return: the path for ``url``.
    :rtype: ``str``
    """
    return urlsplit(url).path[1:].replace("/", "__")



# coding: utf-8

"""

    General Support Tools
    ~~~~~~~~~~~~~~~~~~~~~

"""
import os
import re
import sys
import numpy as np
import pandas as pd
from itertools import chain


if sys.version_info.major == 3 and sys.version_info.minor >= 4:
    from html import unescape
else:
    from six.moves.html_parser import HTMLParser
    unescape = HTMLParser().unescape


try:
    import pydicom as dicom
except ImportError:
    import dicom


try:
    __IPYTHON__
    IN_NOTEBOOK = True
    from IPython.display import display as ipython_display
    from tqdm import tqdm_notebook as tqdm
except NameError:
    IN_NOTEBOOK = False
    ipython_display = None
    from tqdm import tqdm


_only_numeric_regex = re.compile(r'[^\d.]+')
_cln_extent_1_regex = re.compile('\s\s+')
_cln_extent_2_regex = re.compile('\s+')


class InsufficientNumberOfFiles(Exception):
    pass


def isclose(a, b, rel_tol=1e-09, abs_tol=0.0):
    """

    Test if ``a`` is close to ``b``.
    See: https://www.python.org/dev/peps/pep-0485/#proposed-implementation.

    """
    # This is implemented in the `math` module for Python >= 3.5.
    return abs(a-b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)


def dict_reverse(d):
    """

    Reverse a dict.

    :param d:
    :return:
    """
    return {v: k for k, v in d.items()}


def reverse_comma(s):
    """
    
    Reverse a string with a comma.
    
    Note: ``s`` is expected to be passed through ``cln`` first.
    
    :param s: a string
    :type s: ``str``
    :return: reversed string
    :rtype: ``str``
    """
    if isinstance(s, str) and s.count(",") == 1:
        return " ".join(s.split(", ")[::-1])
    else:
        return s


def order_set(l):
    """

    Reduce ``l`` to uniques while preserving order.

    :param l: an iterable
    :type l: ``iterable``
    :return: see description.
    """
    to_return = list()
    for i in l:
        if i not in to_return:
            to_return.append(i)
    return to_return


def is_int(i):
    """

    Checks if the input (``i``) is an integer in a way which is robust against booleans.

    :param i: any input
    :type i: ``any``
    :return: whether or not the input is an integer.
    :rtype: ``bool``
    """
    # Note: ``isinstance(fuzzy_threshold, bool)`` blocks `False` being evaluated as 0 (an intiger).
    return isinstance(i, int) and not isinstance(i, bool)


def is_numeric(i):
    """

    Checks if the input (``i``) is an numeric (``int`` or ``float``)
    in a way which is robust against booleans.

    :param i: any input
    :type i: ``any``
    :return: whether or not the input is an integer.
    :rtype: ``bool``
    """
    return isinstance(i, (float, int)) and not isinstance(i, bool)


def multi_replace(s, to_replace, replace_with=""):
    """

    Run a replace against ``s`` for every item in ``to_replace``.

    :param s: any string
    :type s: ``str``
    :param to_replace: strings to replace in ``s``
    :type to_replace: ``tuple`` or ``list``
    :param replace_with: to replace items in ``to_replace`` with.
    :type replace_with: ``str``
    :return: ``s`` with all items in ``to_replace`` replaced with "".
    :type: ``str``
    """
    for tr in to_replace:
        s = s.replace(tr, replace_with)
    return s


def remove_line_breaks(s):
    """

    Removes all line breaks.

    :param s: any string
    :type s: ``str``
    :return: ``s`` without line breaks.
    :rtype: ``str``
    """
    return re.sub(r'[\t\r\n]', ' ', s)


def remove_html_bullet_points(html):
    """

    Cleans ``html`` by removing any bullet points (html entities)
    and replacing them with semi-colons.

    :param html: html as text with bullet points (HTML entity: '&bull;'
    :type html: ``str``
    :return: see description.
    :rtype: ``str``
    """
    no_points = cln(html).replace('\n&bull; ', "; ").replace('&bull; ', "; ")
    return cln(remove_line_breaks(no_points).replace(" ;", "; "))


def n_sub_dirs(dir):
    """

    :param dir: a path
    :type dir: ``str``
    :return: number of subdirectories in the dir
    """
    if not os.path.isdir(dir):
        raise FileNotFoundError("'{0}' is not a directory.".format(dir))
    return len([k for i, j, k in os.walk(dir)]) - 1


def directory_existence_handler(path_, allow_creation=True, verbose=False):
    """

    Create a directory if it does not exist.

    :param path_: a system path
    :type path_: ``str``
    :param allow_creation: if ``True``, create ``path_`` if it does not exist, else raise. Defaults to ``True``.
    :type allow_creation: ``bool``
    :param verbose: if ``True`` print when a directory is created.
    :type verbose: ``bool``
    :return: ``path_``
    :rtype: ``str``
    """
    if not os.path.isdir(path_):
        if allow_creation:
            os.makedirs(path_)
            if verbose:
                print("\nThe following directory has been created:\n\n{0}\n".format(path_))
        else:
            raise NotADirectoryError("No such directory:\n'{0}'\n".format(path_))
    return path_


def pstr(s):
    """

    Convert to any obect to a string using pandas.
    Source: https://github.com/TariqAHassan/EasyMoney

    :param s: item to be converted to a string.
    :type s: ``any``
    :return: a string
    :rtype: ``str``
    """
    return pd.Series([s]).astype('unicode')[0]


def items_null(element):
    """

    Check if an object is a NaN, including all the elements in an iterable.
    Source: https://github.com/TariqAHassan/EasyMoney

    :param element: a python object.
    :type element: ``any``
    :return: assessment of whether or not `element` is a NaN.
    :rtype: ``bool``
    """
    if isinstance(element, (list, tuple)) or 'ndarray' in str(type(element)):
        return True if all(pd.isnull(i) for i in element) else False
    else:
        return pd.isnull(element)


def list_to_bulletpoints(l, sort_elements=True):
    """

    Convert a list to bullet points.

    :param l: a list (in the colloquial sense) of strings.
    :type l: ``list`` or ``tuple``
    :param sort_elements: if ``True``, sort the elements in the list. Defaults to ``True``.
    :type sort_elements: ``bool``
    :return: list itmes formatted as a string of bullet points (with line breaks).
    :rtype: ``str``
    """
    to_format = sorted(l) if sort_elements else list(l)
    return "".join(map(lambda x: "  - '{0}'\n".format(x), to_format))[:-1]


def header(string, flank=True):
    """

    Generate a Header String

    :param string: a string.
    :type string: ``str``
    :param flank: if True, flank the header with line breaks.
    :type flank: ``bool``
    :return:
    """
    # Compute seperating line
    sep_line = "-" * len(string)

    # Display
    if flank:
        print("\n")
    print("\n{0}\n{1}\n{2}\n".format(sep_line, string, sep_line))
    if flank:
        print("\n")


def camel_to_snake_case(name):
    """

    Source: http://stackoverflow.com/a/1176023/4898004

    :param name:
    :return:
    """
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def cln(i, extent=1, strip=True):
    """

    String white space 'cleaner'.

    :param i: input str
    :type i: ``str``
    :param extent: 1 --> all white space reduced to length 1; 2 --> removal of all white space.
    :param strip: call str.strip()
    :tyoe strip: ``bool``
    :return: cleaned string
    :rtype: ``str``
    """
    to_return = ""
    if isinstance(i, str) and i != "":
        if extent == 1:
            to_return = re.sub(_cln_extent_1_regex, " ", i)
        elif extent == 2:
            to_return = re.sub(_cln_extent_2_regex, "", i)
    else:
        return i

    return to_return.strip() if strip else to_return


def n_split(s, n=2, delim='_'):
    """

    Splits a string into ``n`` groups on a given delimiter.

    Source: http://stackoverflow.com/a/17060409/4898004

    :param s: any string
    :type s: ``str``
    :param n: number of groups. Defaults to `2`.
    :type n: ``int``
    :param delim: a delimiter. Defaults to '_'.
    :type delim: ``str``
    :return: a tuple of with length``n``.
    :rtype: ``tuple``
    """
    groups = s.split(delim)
    return delim.join(groups[:n]), delim.join(groups[n:])


def combine_dicts(dict_a, dict_b):
    """

    :param dict_a:
    :param dict_b:
    :return:
    """
    new = dict_a.copy()
    new.update(dict_b)
    return new


def images_in_dir(dir, return_len=False):
    """

    :param dir:
    :param return_len:
    :return:
    :rtype: ``int`` or ``list``
    """
    if not os.path.isdir(dir):
        raise FileNotFoundError("'{0}' does not exist.".format(str(dir)))
    image_types = (".png", ".jpg", ".tiff", ".gif")
    unnested_list = chain(*[k for i, j, k in os.walk(dir)])
    n_images = [i for i in unnested_list if any(t in i.lower() for t in image_types)]

    return len(n_images) if return_len else n_images


def only_numeric(s, float_output=True):
    """

    Remove all non-numeric characters from a string
    (excluding decimals).

    :param s: a string containing numbers
    :type s: ``str``
    :param float_output: if ``True``, ``float`` the output.
    :type float_output: ``bool``
    :return: the number contained within ``s``.
    :rtype: ``float`` or ``None``
    """
    # See: http://stackoverflow.com/a/947789/4898004
    cleaned = re.sub(_only_numeric_regex, '', s).strip()
    if len(cleaned) and float_output:
        return float(cleaned)
    elif len(cleaned):
        return cleaned
    else:
        return np.NaN


def multimap(data_frame, columns, func):
    """

    Maps some function (``func``) over all columns in ``columns``.

    :param data_frame: any dataframe.
    :type data_frame: ``Pandas DataFrame``
    :param columns: a 'list' of columns.
    :type columns: ``list`` or ``tuple``
    :param func: some function.
    :type func: ``func``
    :return: ``data_frame`` with ``func`` mapped over all ``columns``.
    :rtype: ``Pandas DataFrame``
    """
    if not isinstance(columns, (list, tuple)):
        return data_frame

    for c in columns:
        data_frame[c] = data_frame[c].map(func, na_action='ignore')

    return data_frame


def data_frame_col_drop(data_frame, columns_to_drop, data_frame_name):
    """

    Return ``data_frame`` with select columns dropped.

    :param data_frame: a ``data_frame``
    :type data_frame: ``Pandas DataFrame``
    :param columns_to_drop: a 'list' of columns to drop from ``data_frame``.
    :type columns_to_drop: ``list`` or ``tuple``
    :param data_frame_name: the name of ``data_frame`` to use when raising.
    :type data_frame_name: ``str``
    :return: ``data_frame`` with ``columns_to_drop`` dropped.
    :rtype: ``Pandas DataFrame``
    """
    if not isinstance(data_frame, pd.DataFrame):
        raise TypeError("`{0}` is not a DataFrame.".format(data_frame_name))
    columns_to_keep = [i for i in data_frame.columns if i not in columns_to_drop]
    return data_frame[columns_to_keep]



# coding: utf-8

"""

    Cancer Imaging Archive Parameters Extraction
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
import os
import re
import requests
import pandas as pd
from itertools import chain

# General Support Tools
from biovida.support_tools.support_tools import (cln,
                                                 header,
                                                 n_split,
                                                 combine_dicts,
                                                 camel_to_snake_case)

# Cache Management
from biovida.support_tools._cache_management import package_cache_creator


class CancerImageArchiveParams(object):
    """
    
    :param cache_path: path to the location of the BioVida cache. If a cache does not exist in this location,
                        one will created. Default to ``None``, which will generate a cache in the home folder.
    :type cache_path: ``str`` or ``None``
    :param verbose: if ``True`` print additional information. Defaults to ``False``.
    :type verbose: ``bool``
    """
    
    def __init__(self, cache_path=None, verbose=False,):
        self._verbose = verbose
        
        # Define a location to save the data
        _, self._created_image_dirs = package_cache_creator(sub_dir='images', cache_path=cache_path,
                                                            to_create=['tcia'], nest=[('tcia', 'databases')],
                                                            verbose=verbose)
    
    def _roll_strs_forward(self, l):
        """
    
        Replaces gaps between strings in an list with
        the last string to appear in the list.
    
        :param l: a python list.
        :type l: ``list``
        :return: ``l`` with strings in the list rolled forward (i.e., replacing) all non-strings.
        :rtype: ``list``
        """
        current_string = None
        for i in range(len(l)):
            if isinstance(l[i], str):
                current_string = cln(l[i])
            elif current_string is not None:
                l[i] = current_string
        return l
    
    def _contains_all_cols(self, column_names):
        """
    
        Test to see if the header contains all of the required column names.
    
        :param column_names: column names of the dataframe.
        :type column_names: ``Pandas Series``
        :return:
        """
        required_colnames = ("resource", "queryendpoint", "query parameters", "format", "description")
        return all(any(rc in h for rc in required_colnames) for h in column_names.str.lower())
    
    def _extract_on_required(self, q_params):
        """
    
        Parse Strings which contain '(R)'.
    
        :param q_params: a string containing '(R)'.
        :type q_params: ``str``
        :return: parsed string
        :rtype: ``list``
    
        :Example:
        >>> CancerImageArchiveParams()._extract_on_required('Date (R)Collection(R)PatientID')
        ...
        ['Date (R)', 'Collection (R)', 'PatientID']
        """
        req_split = re.split('(\(R\))', cln(q_params))
    
        final = list()
        for i in range(len(req_split)):
            if i != len(req_split) - 1 and req_split[i + 1] == "(R)":
                final.append([req_split[i] + req_split[i + 1]])
            elif req_split[i] != '(R)':
                final.append(req_split[i].split())
    
        return list(chain(*final))
    
    def _query_parameters_parser(self, q_params):
        """
    
        Parse rows in query_parameters column into a list of tuples.
    
        :param q_params: query_parameters
        :type q_params: ``str``
        :return: a list of tuples of the form: ``[('PARAM_1', 'r' or 'o'), 'PARAM_2', 'r' or 'o'),...]``,
                where 'r' denotes a required parameter and 'o' denotes an optional parameter.
        :rtype: ``list``
        """
        if q_params.strip().lower() == 'none':
            return None
        elif "(R)" in q_params.strip():
            parsed_q_params = self._extract_on_required(q_params)
        else:
            parsed_q_params = list(map(lambda i: i.strip(), cln(q_params).split("/")))
    
        # Mark params as required
        return [(p.replace("(R)", "").strip(), 'r') if "(R)" in p else (p, 'o') for p in parsed_q_params]
    
    def _tcia_api_table_from_html(self, api_ref_loc):
        """
    
        Extract the TCIA API Reference Table from the Usage Guide Wiki.
    
        :param api_ref_loc: url to the TCIA API Usage Guide.
        :type api_ref_loc: ``str``
        :return: the TCIA API Reference Table with unrefined column.
        :rtype: ``Pandas DataFrame``
        """
        html = requests.get(api_ref_loc).text
    
        # Extract all tables from the page
        all_tables = pd.read_html(str(html), header=0)
    
        # Keep only those tables with valid headers
        valid_tables = [t for t in all_tables if self._contains_all_cols(t.columns)]
    
        if len(valid_tables) != 1:
            raise AttributeError("Multiple Valid API Reference Tables Found.")
    
        # Extract the api reference table
        api_df = valid_tables[0]
    
        # Clean Column Names
        c_names = list(map(lambda i: camel_to_snake_case(cln(i, extent=2)), api_df.columns))
        api_df.columns = list(map(lambda i: n_split(i, n=2)[0].strip(), c_names))
    
        return api_df
    
    def _reference_table(self, api_ref_loc):
        """
    
        Extract and Parse the API Reference Table from the Cancer Image Archive Usage Guide Wiki.
    
        :param api_ref_loc: URL to the TCIA API Usage Guide.
        :type api_ref_loc: ``str``
        :return: TCIA API Reference Table.
        :rtype: ``Pandas DataFrame``
        """
        # Extract the c
        api_df = self._tcia_api_table_from_html(api_ref_loc)
    
        # Roll the strings in 'Resource' forward.
        api_df['resource'] = self._roll_strs_forward(api_df['resource'].tolist())
    
        # Clean all columns
        for c in api_df.columns:
            api_df[c] = [cln(i) if isinstance(i, str) else i for i in api_df[c].tolist()]
    
        # Remove Rows with no 'query_endpoint'
        api_df = api_df[pd.notnull(api_df['query_endpoint'])].reset_index(drop=True)
    
        # Parse the 'format' column
        api_df['format'] = api_df['format'].map(lambda x: x.lower().split("/"), na_action='ignore')
    
        # Parse the 'query_parameters' column
        api_df['query_parameters'] = api_df['query_parameters'].map(self._query_parameters_parser)
    
        return api_df
    
    def _reference_table_as_dict(self, api_df):
        """
    
        Return a nested dict of the Cancer Image Archive API Reference table.
    
        :param api_df: Cancer Image Archive API Reference table (dataframe)
        :type api_df: ``Pandas DataFrame``
        :return: dictionary of the form:
    
                ``{'query_endpoint': {'resource', ...'query_parameters': ..., 'format': ..., 'description': ...}, ...}``
    
        :rtype: ``dict``
        """
        nested_dict = dict()
        for (r, qe, qp, f, d) in zip(*[api_df[c].tolist() for c in api_df.columns]):
            row_data = {"resource": r, "query_parameters": qp, "format": f, "description": d}
            if r not in nested_dict:
                nested_dict[qe] = row_data
            else:
                nested_dict[qe] = combine_dicts(nested_dict[qe], row_data)
    
        return nested_dict
    
    def _dicom_long_rename(self):
        """
        
        Replacement 'long' modality names.
        
        :return: a dictionary of the form: ``{current name: updated name}``.
        :rtype: ``dict``
        """
        rename_dict = {
            'Positron emission tomography (PET)': 'Positron Emission Tomography (PET)',
            'Magnetic Resonance': 'Magnetic Resonance Imaging (MRI)',
            'Computed Tomography': 'Computed Tomography (CT)'
        }
        return rename_dict
    
    def dicom_modality_abbreviations(self,
                                     rtype='dataframe',
                                     download_override=False,
                                     modality_loc='https://wiki.cancerimagingarchive.net/display/Public/'
                                                   'DICOM+Modality+Abbreviations'):
        """
        
        Download a dicom Modality Table.

        :param rtype: 'dataframe' for a Pandas DataFrame or 'dict' for a dictionary of the form:

            ``{'short_1': 'long_1', 'short_2': 'long_2', ...}``

        :param rtype: ``str``
        :param download_override: If ``True``, override any existing database currently cached and download a new one.
                                  Defaults to ``False``.
        :type download_override: ``bool``
        :param modality_loc: the location of the DICOM Moldaility Table
        :type modality_loc: ``str``
        :return: DICOM Modality Table
        :rtype: ``Pandas DataFrame``
        """
        if rtype not in ('dataframe', 'dict'):
            raise ValueError("`rtype` must be either 'dataframe' or 'dict'.")

        # Define the path to save the data
        save_path = os.path.join(self._created_image_dirs['databases'], 'tcia_api_dicom_modality.p')

        if not os.path.isfile(save_path) or download_override:
            if self._verbose:
                header("Downloading DICOM Modality Table... ", flank=False)
            html = requests.get(modality_loc).text
            modality_df = pd.read_html(str(html), header=0)[0]
            modality_df.columns = modality_df.columns.str.lower()
            modality_df['long'] = modality_df['long'].replace(self._dicom_long_rename())
            modality_df.to_pickle(save_path)
        else:
            modality_df = pd.read_pickle(save_path)

        return modality_df if rtype == 'dataframe' else dict(zip(modality_df['short'], modality_df['long']))

    def cancer_image_api_ref(self,
                             rtype='dataframe',
                             download_override=False,
                             api_ref_loc='https://wiki.cancerimagingarchive.net/display/Public/'
                                         'TCIA+Programmatic+Interface+%28REST+API%29+Usage+Guide'):
        """
    
        Extracts the API reference for The Cancer Imaging Archive.
    
        :param rtype: 'dataframe' for a Pandas DataFrame or 'dict' for a nested dictionary of the form:
    
            ``{'query_endpoint': {'resource', ..., 'query_parameters': ..., 'format': ..., 'description': ...}, ...}``
    
        :param rtype: ``str``
        :param download_override: If ``True``, override any existing database currently cached and download a new one.
                                  Defaults to ``False``.
        :type download_override: ``bool``
        :param api_ref_loc: URL to the TCIA API Usage Guide.
        :type api_ref_loc: ``str``
        :return:  Cancer Imaging Archive API reference.
        :rtype: ``dict`` or ``Pandas DataFrame``
        """
        if rtype not in ('dataframe', 'dict'):
            raise ValueError("`rtype` must be either 'dataframe' or 'dict'.")
    
        # Define the path to save the data
        save_path = os.path.join(self._created_image_dirs['databases'], 'tcia_api_reference.p')
    
        if not os.path.isfile(save_path) or download_override:
            if self._verbose:
                header("Downloading API Reference Table... ", flank=False)
            api_reference = self._reference_table(api_ref_loc)
            api_reference.to_pickle(save_path)
        else:
            api_reference = pd.read_pickle(save_path)
    
        # Return based on `rtype`.
        return api_reference if rtype == 'dataframe' else self._reference_table_as_dict(api_reference)



# coding: utf-8

"""

    Disease-Symptoms Interface
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
import os
import pandas as pd

# General Support Tools
from biovida.support_tools.support_tools import cln, header
from biovida.support_tools._cache_management import package_cache_creator


class DiseaseSymptomsInterface(object):
    """

    Tools to obtain databases relating symptoms to diseases.

    References:

    - Daniel Himmelstein, Antoine Lizee, Chrissy Hessler, Leo Brueggeman, Sabrina Chen, Dexter Hadley, Ari Green,
      Pouya Khankhanian, Sergio Baranzini. Rephetio: Repurposing drugs on a hetnet [report]. Thinklab (2016).
      doi:`10.15363/thinklab.a7 <http://www.thinklab.com/p/rephetio/report#edges>`_. Online Repository:
      https://github.com/dhimmel/medline.

    - Zhou, XueZhong, Jörg Menche, Albert-László Barabási, and Amitabh Sharma. Human symptoms–disease network.
      Nature communications 5 (2014). doi:`10.1038/ncomms5212 <http://www.nature.com/articles/ncomms5212>`_.
      Online Repository: https://github.com/dhimmel/hsdn.

    :param cache_path: location of the BioVida cache. If one does not exist in this location, one will created.
                       Default to ``None`` (which will generate a cache in the home folder).
    :type cache_path: ``str`` or ``None``
    :param verbose: If ``True``, print notice when downloading database. Defaults to ``True``.
    :type verbose: ``bool``
    """

    def __init__(self, cache_path=None, verbose=True):
        self._verbose = verbose
        # Cache creation
        pcc = package_cache_creator(sub_dir='diagnostics',
                                    to_create=['disease_symptoms'],
                                    cache_path=cache_path,
                                    verbose=verbose)
        self.root_path, self._created_disease_ont_dirs = pcc

        # URLs to the databases
        self._hsdn_url = "https://raw.githubusercontent.com/LABrueggs/HSDN/master/Combined-Output.tsv"
        self._rephetio_ml_url = "https://github.com/dhimmel/medline/raw/0c9e2905ccf8aae00af5217255826fe46cba3d30/data/disease-symptom-cooccurrence.tsv"

        self.hsdn_db = None
        self.rephetio_ml_db = None
        self.combined_db = None

    @staticmethod
    def _clean_columns(columns):
        """

        Convert ``columns`` into snake_case.

        :param columns: a list of dataframe columns
        :type columns: ``Pandas Series`` or ``list``
        :return: columns in snake_case
        :rtype: ``list``
        """
        return list(map(lambda x: cln(x).replace(" ", "_").lower(), list(columns)))

    @staticmethod
    def _comma_reverse(term):
        """

        Reverse a single comma, e.g., 'car, blue' --> 'blue car'.

        :param term: a string
        :type term: ``str``
        :return: ``term`` with the comma reversed, if present.
        :rtype: ``str``
        """
        if not isinstance(term, str):
            return term
        elif term.count(", ") != 1:
            return cln(term)
        else:
            return " ".join(cln(term).split(", ")[::-1])

    @staticmethod
    def _harvest(url, cleaner_func, save_path, download_override):
        """

        Pull (i.e., download) a Database.

        :param download_override: If ``True``, override any existing database currently cached and download a new one.
        :type download_override: ``bool``
        :return: the Human Symptoms Disease Network database as a DataFrame.
        :rtype: ``Pandas DataFrame``
        """
        if not os.path.isfile(save_path) or download_override:
            to_return = pd.read_csv(url, sep="\t", header=0, index_col=0).reset_index(drop=True)
            to_return = cleaner_func(to_return)
            to_return.to_pickle(save_path)
        else:
            to_return = pd.read_pickle(save_path)

        return to_return

    def _hsdn_df_cleaner(self, data_frame):
        """

        Clean the harvested HSDN Database in two ways:

        - Lower convert column names to snake_case and lowercase

        - Reverse commas, e.g., 'Arthritis, Rheumatoid' --> 'Rheumatoid Arthritis' (column name: 'common_disease_name').

        :param data_frame: the dataframe obtained in the ``hsdn_pull()`` method.
        :type data_frame: ``Pandas DataFrame``
        """
        # Correct the column names
        data_frame.columns = self._clean_columns(data_frame.columns)

        # Reverse commas
        data_frame['common_disease_name'] = data_frame['mesh_disease_term'].map(self._comma_reverse)
        data_frame['common_symptom_term'] = data_frame['mesh_symptom_term'].map(self._comma_reverse)

        return data_frame

    def hsdn_pull(self, download_override=False):
        """

        Pull (i.e., download) the Human Symptoms Disease Network Database.

        :param download_override: If ``True``, override any existing database currently cached and download a new one.
        :type download_override: ``bool``
        :return: the Human Symptoms Disease Network database as a DataFrame.
        :rtype: ``Pandas DataFrame``
        """
        save_path = os.path.join(self._created_disease_ont_dirs['disease_symptoms'], 'hsdn_db.p')

        if self._verbose:
            header("Downloading Human Symptoms Disease Network Database... ", flank=False)

        # Harvest the database
        self.hsdn_db = self._harvest(url=self._hsdn_url,
                                     cleaner_func=self._hsdn_df_cleaner,
                                     save_path=save_path,
                                     download_override=download_override)

        return self.hsdn_db

    def _rephetio_ml_df_cleaner(self, data_frame):
        """

        Clean the harvested Rephetio Medline Database in two ways:

        - Lower convert column names to snake_case and lowercase

        - Reverse commas, e.g., 'Arthritis, Rheumatoid' --> 'Rheumatoid Arthritis' (column name: 'common_disease_name').

        :param data_frame: the dataframe obtained in the ``rephetio_ml_pull()`` method.
        :type data_frame: ``Pandas DataFrame``
        :return: a cleaned ``data_frame``
        :rtype: ``Pandas DataFrame``
        """
        # Correct the column names
        data_frame.columns = self._clean_columns(data_frame.columns)

        # Reverse commas
        data_frame['common_symptom_term'] = data_frame['mesh_name'].map(self._comma_reverse)

        return data_frame

    def rephetio_ml_pull(self, download_override=False):
        """

        Pull (i.e., download) the Rephetio Medline Database.

        :param download_override: If ``True``, override any existing database currently cached and download a new one.
        :type download_override: ``bool``
        :return: the Rephetio Medline database as a DataFrame.
        :rtype: ``Pandas DataFrame``
        """
        save_path = os.path.join(self._created_disease_ont_dirs['disease_symptoms'], 'rephetio_medline_db.p')

        if self._verbose:
            header("Downloading Rephetio Medline Database... ", flank=False)

        # Harvest the database
        self.rephetio_ml_db = self._harvest(url=self._rephetio_ml_url,
                                            cleaner_func=self._rephetio_ml_df_cleaner,
                                            save_path=save_path,
                                            download_override=download_override)

        return self.rephetio_ml_db

    def _combine(self, download_override, **kwargs):
        """

        Combine `Rephetio Medline` and `Human Symptoms Disease Network` databases.

        :param download_override: If ``True``, override any existing database currently cached and download a new one.
        :type download_override: ``bool``
        :return: a combined dataframe with the following columns: 'common_disease_name' and 'common_symptom_term'.
        :rtype: ``Pandas DataFrame``
        """
        if 'hsdn' in kwargs:
            hsdn = kwargs.get('hsdn')
        else:
            hsdn = self.hsdn_pull(download_override).copy(deep=True)

        if 'rephetio' in kwargs:
            rephetio = kwargs.get('rephetio')
        else:
            rephetio = self.rephetio_ml_pull(download_override).copy(deep=True)

        # Rename `rephetio` columns to s.t. they match with `hsdn`.
        rephetio = rephetio.rename(columns={'doid_name': 'common_disease_name'})

        if self._verbose:
            header("Combining Databases... ", flank=False)

        concat_on_cols = ['common_disease_name', 'common_symptom_term']
        combined_db = pd.concat([hsdn[concat_on_cols], rephetio[concat_on_cols]], ignore_index=True)

        return combined_db.drop_duplicates().reset_index(drop=True)

    def pull(self, download_override=False):
        """

        Construct a dataframe by combining the `Rephetio Medline` and `Human Symptoms Disease Network` databases.

        :param download_override: If ``True``, override any existing database currently cached and download a new one.
        :type download_override: ``bool``
        :return: a combined dataframe with the following columns: 'common_disease_name' and 'common_symptom_term'.
        :rtype: ``Pandas DataFrame``
        """
        save_path = os.path.join(self._created_disease_ont_dirs['disease_symptoms'], 'combined_ds_db.p')

        if not os.path.isfile(save_path) or download_override:
            self.combined_db = self._combine(download_override)
            self.combined_db.to_pickle(save_path)
        else:
            self.combined_db = pd.read_pickle(save_path)

        return self.combined_db
